@@ -54,20 +54,17 @@ bool ProfileCreator::CreateProfile(const string &input_profile_name,
 bool ProfileCreator::ReadSample(const string &input_profile_name,
                                 const string &profiler) {
   if (profiler == "perf") {
-    // Sets the regular expression to filter samples for a given binary.
-    char *dup_name = strdup(binary_.c_str());
-    char *strip_ptr = strstr(dup_name, ".unstripped");
-    if (strip_ptr) {
-      *strip_ptr = 0;
+    string file_base_name = basename(binary_.c_str());
+    size_t unstripped_at = file_base_name.find(".unstripped");
+    if (unstripped_at != string::npos) {
+      file_base_name.erase(unstripped_at);
     }
-    const char *file_base_name = basename(dup_name);
-    CHECK(file_base_name) << "Cannot find basename for: " << binary_;
+    CHECK(!file_base_name.empty()) << "Cannot find basename for: " << binary_;
 
     ElfReader reader(binary_);
 
     sample_reader_ = new PerfDataSampleReader(
-        input_profile_name, file_base_name);
-    free(dup_name);
+        input_profile_name, std::move(file_base_name));
   } else if (profiler == "text") {
     sample_reader_ = new TextSampleReaderWriter(input_profile_name);
   } else {
@@ -83,8 +80,8 @@ bool ProfileCreator::ReadSample(const string &input_profile_name,
 
 bool ProfileCreator::ComputeProfile(SymbolMap *symbol_map,
                                     Addr2line **addr2line) {
-  set<uint64> sampled_addrs = sample_reader_->GetSampledAddresses();
-  map<uint64, uint64> sampled_functions =
+  std::set<uint64> sampled_addrs = sample_reader_->GetSampledAddresses();
+  std::map<uint64, uint64> sampled_functions =
       symbol_map->GetSampledSymbolStartAddressSizeMap(sampled_addrs);
   *addr2line =
       Addr2line::CreateWithSampledFunctions(binary_, &sampled_functions);
@@ -103,10 +100,11 @@ bool ProfileCreator::ComputeProfile(SymbolMap *symbol_map,
 bool ProfileCreator::CreateProfileFromSample(ProfileWriter *writer,
                                              const string &output_name) {
   SymbolMap symbol_map(binary_);
+  symbol_map.set_use_discriminator_encoding(use_discriminator_encoding_);
   Addr2line *addr2line = nullptr;
   if (!ComputeProfile(&symbol_map, &addr2line)) return false;
 
-  ModuleGrouper *grouper =
+  auto grouper =
       ModuleGrouper::GroupModule(binary_, GCOV_ELF_SECTION_NAME, &symbol_map);
 
   writer->setSymbolMap(&symbol_map);
@@ -114,7 +112,6 @@ bool ProfileCreator::CreateProfileFromSample(ProfileWriter *writer,
   bool ret = writer->WriteToFile(output_name);
 
   delete addr2line;
-  delete grouper;
   return ret;
 }
 
